// ==============================================================
//                 ORBITER MODULE: danuri
//                  
//
// danuri.cpp
// Control module for danuri vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double ISP = 2227;
const double RCSISP = 2139;
const double RCSTH = 3.48;
const double MAINTH = 127.2;
const double ARRAY_SPEED = 0.2;
const double ARRAY_ROT_SPEED = 0.1;
const double BOOM_SPEED = 0.5;
const double ANT_SPEED = 0.1;
const double FUEL_MASS = 260;
const double EMP_MASS = 418;

#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)

static TOUCHDOWNVTX tdvtx_geardown[3] = {
	{_V( 0,-1, -1.32), 6e4, 2e3, 3.0, 0.2},
	{_V(-0.5, 0.866, -1.32), 6e4, 2e3, 3.0, 0.2},
	{_V( 0.5, 0.866, -1.32), 6e4, 2e3, 3.0, 0.2}
};




// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
//void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
//{
//	int i;
//	const int nabsc = 9;
//	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
//	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
//	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
//	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
//	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
//	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
//	*cm = 0.0;
//	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
//}

//void HLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
//{
//	int i;
//	const int nabsc = 9;
//	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
//	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
//	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
//	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
//	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
//	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
//	*cm = 0.0;
//	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
//}


class danuri: public VESSEL3 {
public:
	
	danuri (OBJHANDLE hVessel, int flightmodel);
	~danuri ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	

private:
	void DefineAnimations (void);
	void DoKillRot(void);
//	void DoPrograde(void);
//	void DoRetrograde(void);
//	void DoOrbitNormal(void);
//	void DoOrbitAntiNormal(void);
	void DoNadirPointing(void);

	MGROUP_TRANSFORM *MgrpTran[8];
	UINT anim_array_deploy, anim_array_rot, anim_ant_roty, anim_ant_rotx, anim_boom;
//	int MyID;
	THRUSTER_HANDLE th_main[4], th_rcs[8], th_group[4];
	PROPELLANT_HANDLE hProp;
	ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2;
	//ATTACHMENTHANDLE hAttach;
	double array_deploy_proc, array_rot_proc, ant_rotx_proc, ant_roty_proc, boom_proc;
	int array_status, ant_status, boom_status, kilrot_sw, nadir_sw;


};

danuri::danuri (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{

}

danuri::~danuri ()
{
int i;
for (i = 0; i < 8; i++)
	{
	delete MgrpTran[i];
	}
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void danuri::DefineAnimations (void)
{
	// array
	anim_array_rot = CreateAnimation (0);

	static UINT ArrayGrp0[1] = {19};
	MgrpTran[0] = new MGROUP_ROTATE(0, ArrayGrp0, 0, _V(0,0,0), _V(1,0,0), (float)(PI2));
	Section0 = AddAnimationComponent (anim_array_rot, 0.0, 1.0, MgrpTran[0]);

	anim_array_deploy = CreateAnimation (0);

	//port
	static UINT ArrayGrp1[1] = {19};
	MgrpTran[1] = new MGROUP_ROTATE(0, ArrayGrp1, 1, _V(-0.89,0,0), _V(0,1,0), (float)(90*RAD));
	Section1 = AddAnimationComponent (anim_array_deploy, 0.0, 1.0, MgrpTran[1], Section0);

	static UINT ArrayGrp2[1] = {0};
	MgrpTran[2] = new MGROUP_ROTATE(0, ArrayGrp2, 1, _V(-0.903,0,-0.886), _V(0,1,0), (float)(-180*RAD));
	AddAnimationComponent (anim_array_deploy, 0.0, 1.0, MgrpTran[2], Section1);

	// stbd
	static UINT ArrayGrp3[1] = {17};
	MgrpTran[3] = new MGROUP_ROTATE(0, ArrayGrp3, 1, _V(0.89,0,0), _V(0,1,0), (float)(-90*RAD));
	Section2 = AddAnimationComponent (anim_array_deploy, 0.0, 1.0, MgrpTran[3], Section0);

	static UINT ArrayGrp4[1] = {18};
	MgrpTran[4] = new MGROUP_ROTATE(0, ArrayGrp4, 1, _V(0.903,0,-0.886), _V(0,1,0), (float)(180*RAD));
	AddAnimationComponent (anim_array_deploy, 0.0, 1.0, MgrpTran[4], Section2);

	///////////////////// mag boom

	anim_boom = CreateAnimation (0);

	static UINT BoomGrp0[1] = {16};
	MgrpTran[5] = new MGROUP_ROTATE(0, BoomGrp0, 1, _V(0,-0.91, 0.977), _V(1,0,0), (float)(135*RAD));
	AddAnimationComponent (anim_boom, 0.0, 1.0, MgrpTran[5]);


	///////////////////// hga

	anim_ant_roty = CreateAnimation (0);

	static UINT AntGrp0[1] = {3};
	MgrpTran[6] = new MGROUP_ROTATE(0, AntGrp0, 0, _V(0, 0, 0.539), _V(0,1,0), (float)(PI2));
	Section0 = AddAnimationComponent (anim_ant_roty, 0.0, 1.0, MgrpTran[6]);

	anim_ant_rotx = CreateAnimation (0);

	static UINT AntGrp1[3] = {3,4,5};
	MgrpTran[7] = new MGROUP_ROTATE(0, AntGrp1, 3, _V(0, 1.362, 0.539), _V(1,0,0), (float)(PI05));
	AddAnimationComponent (anim_ant_rotx, 0.0, 1.0, MgrpTran[7], Section0);

}



// ==============================================================
// danuri class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void danuri::clbkSetClassCaps (FILEHANDLE cfg)
{
	array_deploy_proc = 0.0;
	array_rot_proc = 0.0;
	array_status = 0;
	ant_rotx_proc = 0;
	ant_roty_proc = 0;
	ant_status = 0;
	boom_proc = 0;
	boom_status = 0;
	kilrot_sw = 0;
	nadir_sw = 0;
	
	DefineAnimations ();

	// physical specs
	SetCW (1, 1, 1, 1);
	SetRotDrag (_V(1,1,1));
	SetPMI (_V(0.6,0.7,0.7));
	SetCrossSections (_V(4,4,4));
	SetSize (1);
	SetEmptyMass (EMP_MASS);
	SetMeshVisibilityMode ((AddMesh ("danuri/danuri")), MESHVIS_EXTERNAL);
	
	SetCameraOffset (_V(0,0,0));
	//SetTouchdownPoints (_V(0.1,0.8,-1.8), _V(0.512,-0.853,-1.8), _V(-0.512,-0.853,-1.8));
	SetTouchdownPoints ( tdvtx_geardown, 3);
	//SetDockParams (_V(0,0,-2), _V(0,0,-1), _V(0,1,0));

	// propellant resources

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************

	// rcs +z + 15deg
	th_rcs[0] = CreateThruster (_V( 0.868, 0.698,-1.324), _V(0.259, 0, 0.966), RCSTH, hProp, RCSISP);
	th_rcs[1] = CreateThruster (_V(-0.868, 0.698,-1.324), _V(-0.259, 0, 0.966), RCSTH, hProp, RCSISP);
	th_rcs[2] = CreateThruster (_V( 0.868,-0.698,-1.324), _V( 0.259, 0, 0.966), RCSTH, hProp, RCSISP);
	th_rcs[3] = CreateThruster (_V(-0.868,-0.698,-1.324), _V(-0.259, 0, 0.966), RCSTH, hProp, RCSISP);

	// Y axis
	th_rcs[4] = CreateThruster (_V( 0.874, 0.802,-1.300), _V( 0,-1, 0), RCSTH, hProp, RCSISP);
	th_rcs[5] = CreateThruster (_V(-0.874, 0.802,-1.300), _V( 0,-1, 0), RCSTH, hProp, RCSISP);
	th_rcs[6] = CreateThruster (_V( 0.874,-0.802,-1.300), _V( 0, 1, 0), RCSTH, hProp, RCSISP);
	th_rcs[7] = CreateThruster (_V(-0.874,-0.802,-1.300), _V( 0, 1, 0), RCSTH, hProp, RCSISP);

	th_main[0] = CreateThruster (_V(-0.125, 0.125,-1.306), _V(0, 0, 1), MAINTH, hProp, ISP);
	th_main[1] = CreateThruster (_V( 0.125, 0.125,-1.306), _V(0, 0, 1), MAINTH, hProp, ISP);
	th_main[2] = CreateThruster (_V(-0.125,-0.125,-1.306), _V(0, 0, 1), MAINTH, hProp, ISP);
	th_main[3] = CreateThruster (_V( 0.125,-0.125,-1.306), _V(0, 0, 1), MAINTH, hProp, ISP);

	

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 5];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 6];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 4];
	th_group[1] = th_rcs[ 7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[ 5];
	th_group[1] = th_rcs[ 6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	th_group[2] = th_rcs[ 2];
	th_group[3] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_main[ 0];
	th_group[1] = th_main[ 1];
	th_group[2] = th_main[ 2];
	th_group[3] = th_main[ 3];
	CreateThrusterGroup (th_group, 4, THGROUP_MAIN);

	///////////////////////////////////////

	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	
	double ex_lng = 0.6;
	double ex_wdt = 0.03;
	AddExhaust (th_rcs[0], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[1], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[2], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[3], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[4], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[5], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[6], ex_lng, ex_wdt, (tex_rcs));
	AddExhaust (th_rcs[7], ex_lng, ex_wdt, (tex_rcs));

	SURFHANDLE tex_main = oapiRegisterExhaustTexture ("exhaust");

	double main_ex_lng = 1.50;
	double main_ex_wdt = 0.06;
	AddExhaust (th_main[0], main_ex_lng, main_ex_wdt, 0.02, (tex_main));
	AddExhaust (th_main[1], main_ex_lng, main_ex_wdt, 0.02, (tex_main));
	AddExhaust (th_main[2], main_ex_lng, main_ex_wdt, 0.02, (tex_main));
	AddExhaust (th_main[3], main_ex_lng, main_ex_wdt, 0.02, (tex_main));


	///////////////////////////////////////////////////////
	/////AirFoils

//	CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0), VLiftCoeff, 0, 1, 1, 1);
//	CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0), HLiftCoeff, 0, 1, 1, 1);

	//////////////////// attach //////////////////////////

//	hAttach = CreateAttachment (true, _V( 0, 0, -2.39), _V( 0,0,-1), _V(0,1,0), "F9");

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new danuri (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (danuri*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void danuri::clbkPostCreation ()
{
	
	/*MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\Vessel\\attsustain.wav",	REPLACE_MAIN_THRUST);
	SoundOptionOnOff(MyID,PLAYGPWS,FALSE);
	SoundOptionOnOff(MyID,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff(MyID,PLAYLANDINGANDGROUNDSOUND,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINAIRCONDITIONING,FALSE);
	SoundOptionOnOff(MyID,PLAYCABINRANDOMAMBIANCE,FALSE);
	SoundOptionOnOff(MyID,DISPLAYTIMER,FALSE);
	SoundOptionOnOff(MyID,PLAYDOCKLANDCLEARANCE,FALSE);
	SoundOptionOnOff(MyID,PLAYRADARBIP,FALSE);*/
	
}
void danuri::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters

	sprintf (cbuf, "%d %0.4f %0.4f", array_status, array_deploy_proc, array_rot_proc);
	oapiWriteScenario_string (scn, "SOLP", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f", ant_status, ant_roty_proc, ant_rotx_proc);
	oapiWriteScenario_string (scn, "ANTN", cbuf);

	sprintf (cbuf, "%d %0.4f", boom_status, boom_proc);
	oapiWriteScenario_string (scn, "BOOM", cbuf);

	sprintf (cbuf, "%d", kilrot_sw);
	oapiWriteScenario_string (scn, "KROT", cbuf);

	sprintf (cbuf, "%d", nadir_sw);
	oapiWriteScenario_string (scn, "NADR", cbuf);

	
}
void danuri::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "SOLP", 4))
		{
            sscanf (line+4, "%d%lf%lf", &array_status, &array_deploy_proc, &array_rot_proc);
		}
		else if (!strnicmp (line, "ANTN", 4))
		{
            sscanf (line+4, "%d%lf%lf", &ant_status, &ant_roty_proc, &ant_rotx_proc);
		}
		else if (!strnicmp (line, "BOOM", 4))
		{
            sscanf (line+4, "%d%lf", &boom_status, &boom_proc);
		}
		else if (!strnicmp (line, "KROT", 4))
		{
            sscanf (line+4, "%d", &kilrot_sw);
		}
		else if (!strnicmp (line, "NADR", 4))
		{
            sscanf (line+4, "%d", &nadir_sw);
		}
		else
		{
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	
	SetAnimation (anim_array_deploy, array_deploy_proc);
	SetAnimation (anim_array_rot, array_rot_proc);
	SetAnimation (anim_ant_roty, ant_roty_proc);
	SetAnimation (anim_ant_rotx, ant_rotx_proc);
	SetAnimation (anim_boom, boom_proc);
	
}






int danuri::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{

//bool cbSelectTarget(void *id, char *str, void *data);

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) // +shift
	{
	  switch (key) 
	  {
		case OAPI_KEY_NUMPAD5: 
		if (kilrot_sw  == 0)kilrot_sw = 1;
		else kilrot_sw = 0;
		return 1;
	  }
	  return 0;
	} 
  else if (KEYMOD_CONTROL (kstate)) // +ctrl
	{
	return 0;
	}
  else // unmodified keys
	{ 
    switch (key) 
		{
		case OAPI_KEY_G: // solp and ant
		if (array_status  == 0)array_status = 1;
		if (ant_status  == 0)ant_status = 1;
		return 1;

		case OAPI_KEY_K: // boom
		if (boom_status  == 0)boom_status = 1;
		return 1;

		case OAPI_KEY_N: // nadir pointing
		if (nadir_sw  == 0)nadir_sw = 1;
		else nadir_sw = 0;
		return 1;
		}
	return 0;
	}
  
 return 0;
 }


int danuri::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
return 0;
}





bool danuri::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	//int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	//int w5 = (int)(264*d);
	//int w6 = (int)(280*d);
	//int w7 = (int)(296*d);
	//int w8 = (int)(312*d);
	//int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	{
	skp->SetTextColor(0x0066FF66);
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2)/(m1));
	double dv = (ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Available dV %0.3f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}


	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"Shift+NumPad = Reaction Wheel Attitude Control");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x0066FF66);
	if(kilrot_sw == 1)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"Shift+NumPad5 = Reaction Wheel KillRot");
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	
	{
	skp->SetTextColor(0x0066FF66);
	if(nadir_sw == 1)skp->SetTextColor(0x0066FFFF);
	char abuf[256];
	sprintf(abuf,"N = Auto Nadir Pointing");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}


	if(array_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Deploy Solar Array and HGA");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	if(boom_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"K = Deploy Magnetometer Boom");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	


	return true;
}


void danuri::clbkPreStep(double SimT, double SimDT, double MJD)
{
	
	if (kilrot_sw  == 1)DoKillRot();

	if(nadir_sw == 1)DoNadirPointing();

	///////// animations //////////////////////////////////

	// deploy solp
	if (array_status == 0) array_deploy_proc = 0;
	else if (array_status == 1) 
	{
		double da = SimDT * ARRAY_SPEED;
		array_deploy_proc = array_deploy_proc + da;
		if (array_deploy_proc > 1) 
		{
			array_deploy_proc = 1;
			array_status = 2;
		}
	}
	else if (array_status == 2) array_deploy_proc = 1;
	SetAnimation (anim_array_deploy, array_deploy_proc);

	/////////////////// array sun tracking

	if (array_status == 2)
	{
		VECTOR3 glob_sun_pos = _V(0,0,0);
		VECTOR3 local_sun_pos = _V(0,0,0);
		Global2Local(glob_sun_pos, local_sun_pos);
		local_sun_pos.x = 0;
		normalise(local_sun_pos);
		double sun_angle = atan2(local_sun_pos.y, -local_sun_pos.z);
		if(sun_angle < 0)sun_angle = sun_angle + PI2;
		double solp_angle = array_rot_proc * PI2;

		VECTOR3 solp_dir = _V(0, sin(solp_angle), -cos(solp_angle));

		//double angle_diff = acos(dotp(solp_dir, local_sun_pos));

		double angle_diff = sun_angle - solp_angle;
		if(angle_diff > PI)angle_diff = angle_diff - PI2;
		else if(angle_diff < -PI)angle_diff = angle_diff + PI2;

		VECTOR3 solp_sun_nml = crossp(solp_dir,local_sun_pos);

		int rot_dir = 1;
		if(solp_sun_nml.x < 0)rot_dir = -1;

		double max_drot = SimDT * ARRAY_ROT_SPEED * PI2;
		if(max_drot > abs(angle_diff)) 
		{
			array_rot_proc = sun_angle/PI2;
		//	sprintf(oapiDebugString(), "1 angle_diff %0.4f max_drot %0.4f",angle_diff*DEG, max_drot*DEG);
		}
		else
		{
		array_rot_proc = array_rot_proc + (SimDT * ARRAY_ROT_SPEED * rot_dir);
		int arp_int = (int)(array_rot_proc);
		if (array_rot_proc > 1)array_rot_proc = array_rot_proc - arp_int;
		else if (array_rot_proc < 0) array_rot_proc = array_rot_proc - (arp_int - 1);
		//sprintf(oapiDebugString(), "2 angle_diff %0.4f max_drot %0.4f",angle_diff*DEG, max_drot*DEG);
		}

		SetAnimation (anim_array_rot, array_rot_proc);
	}

	//////////// mag boom

	if (boom_status == 0) boom_proc = 0;
	else if (boom_status == 1) 
	{
		double da = SimDT * BOOM_SPEED;
		boom_proc = boom_proc + da;
		if (boom_proc > 1) 
		{
			boom_proc = 1;
			boom_status = 2;
		}
	}
	else if (boom_status == 2) boom_proc = 1;
	SetAnimation (anim_boom, boom_proc);

	//////////////////////////////// HGA

	if(ant_status == 1)
	{
	VECTOR3 earth_glob_pos = _V(0,0,0);
	VECTOR3 earth_loc_pos = _V(0,-1,0);
	OBJHANDLE hEarth = oapiGetObjectByName("Earth");
	if (hEarth)
	{
		oapiGetGlobalPos(hEarth, &earth_glob_pos);
		Global2Local(earth_glob_pos, earth_loc_pos);
	}
	normalise(earth_loc_pos);

	double earth_y_angle = atan2(-earth_loc_pos.x, -earth_loc_pos.z);
	if(earth_y_angle < 0)earth_y_angle = earth_y_angle + PI2;

	double ant_y_angle = ant_roty_proc * PI2;

	double y_angle_diff = earth_y_angle - ant_y_angle;
	if(y_angle_diff > PI)y_angle_diff = y_angle_diff - PI2;
	else if(y_angle_diff < -PI)y_angle_diff = y_angle_diff + PI2;

	//VECTOR3 earth_loc_zx = earth_loc_pos;
	//earth_loc_zx.y = 0;
	//normalise(earth_loc_zx);

	//VECTOR3 ant_zx = _V(sin(ant_y_angle),0,-cos(ant_y_angle));
	//normalise(ant_zx);

	//VECTOR3 anty_earth_nml = crossp(ant_zx, earth_loc_zx);
	//normalise(anty_earth_nml);
	//sprintf(oapiDebugString(), "x %0.4f y %0.4f z %0.4f",anty_earth_nml.x, anty_earth_nml.y, anty_earth_nml.z);

	int roty_dir = 1;
	//if(anty_earth_nml.y < 0)roty_dir = -1;
	if(y_angle_diff < 0) roty_dir = -1;

	double max_droty = SimDT * ANT_SPEED * PI2;

	if(max_droty > abs(y_angle_diff))ant_roty_proc = earth_y_angle/PI2;
	else
		{
			double da = SimDT * ANT_SPEED * roty_dir;
			ant_roty_proc = ant_roty_proc + da;
			
			int aryp = (int) (ant_roty_proc);
			if(ant_roty_proc > 1) ant_roty_proc = ant_roty_proc - aryp;
			else if(ant_roty_proc < 0) ant_roty_proc = ant_roty_proc - aryp + 1;
		}
		//sprintf(oapiDebugString(), "roty_dir %d y_angle_diff %0.4f earth_y_angle %0.4f ant_y_angle %0.4f",roty_dir, y_angle_diff*DEG, earth_y_angle*DEG, ant_y_angle*DEG);
		SetAnimation (anim_ant_roty, ant_roty_proc);

		//////////////////////////// xrot

		double cos_yaxis_earth_angle = dotp(_V(0,1,0), earth_loc_pos);
		if(cos_yaxis_earth_angle > 1)cos_yaxis_earth_angle = 1;
		else if(cos_yaxis_earth_angle < 0)cos_yaxis_earth_angle = 0;
		double yaxis_earth_angle = acos(cos_yaxis_earth_angle);
		yaxis_earth_angle = PI05 - yaxis_earth_angle;

		double ant_rotx_angle = ant_rotx_proc * PI05;
		double xrot_angle_diff = yaxis_earth_angle - ant_rotx_angle;

		double max_drotx = SimDT * ANT_SPEED;
		if(max_drotx > abs(xrot_angle_diff)) ant_rotx_proc = yaxis_earth_angle/PI05;
		else
		{
			int rotx_dir = 1;
			if(xrot_angle_diff < 0) rotx_dir = -1;
			double dx = SimDT * ANT_SPEED * rotx_dir;
			ant_rotx_proc = ant_rotx_proc + dx;
			if(ant_rotx_proc > 1) ant_rotx_proc = 1;
			else if(ant_rotx_proc < 0) ant_rotx_proc = 0;
		}
		

		SetAnimation (anim_ant_rotx, ant_rotx_proc);

	}

	
}


//void danuri::clbkPostStep (double simt, double simdt, double mjd)
//{
//
//}

void danuri::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) kilrot_sw = 0;
}



void danuri::DoNadirPointing()
{
	VESSELSTATUS vs;
	GetStatus(vs);

	//OBJHANDLE hGbod = GetGravityRef();
	OBJHANDLE hGbod = oapiGetObjectByName("Moon");

	VECTOR3 pos_unit = _V(0,0,0);
	GetRelativePos(hGbod, pos_unit);
	normalise(pos_unit);

	VECTOR3 vel_unit = vs.rvel;
	normalise(vel_unit);

	VECTOR3 my_pos_glob_unit = _V(0,0,0);
	GetGlobalPos(my_pos_glob_unit);
	normalise(my_pos_glob_unit);

	//VECTOR3 vel_pos_nml = crossp(vel_unit, pos_unit);
	//normalise(vel_pos_nml);

	VECTOR3 x_unit = crossp(pos_unit, my_pos_glob_unit);
	normalise(x_unit);

	VECTOR3 y_unit = pos_unit;
	normalise(y_unit);

	VECTOR3 z_unit = crossp(x_unit,y_unit);
	normalise(z_unit);

	//sprintf(oapiDebugString(), "x %0.3f y %0.3f z %0.3f tot %0.3f",x_unit.x, x_unit.y, x_unit.z, length(x_unit));


	VECTOR3 xright = x_unit;
	VECTOR3 yup = y_unit;
	VECTOR3 zforward = z_unit;


double gamma = atan2( xright.x, yup.x);
double beta = -1*(asin(zforward.x));
double alpha = atan2(zforward.y, zforward.z);

vs.arot = _V(alpha, beta, (PI/2)-gamma);
vs.vrot = _V(0,0,0);

DefSetState(&vs);

}